# Copyright (c) 2021 J. Lewis Muir
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require 'csv'
require 'fileutils'
require 'optparse'

require_relative 'application'

#
# Reads and writes IMCA-CAT SynchWeb shipment CSV files.
#
class IMCAShipmentToolApp < Application
  class Error < StandardError
  end

  IOPair = Struct.new(:in, :out)
  SortOptions = Struct.new(:help, :in_place, :verbose, :ref_file)

  CSV_PIN_COLUMN_NAME = 'Pin'.freeze

  # Converts SystemCallError to Error with pathname-first message.
  def self.open_file(pathname, *args, &block)
    File.open(pathname, *args, &block)
  rescue SystemCallError => e
    raise Error.new("#{pathname}: #{SystemCallError.new(e.errno)}")
  end

  # Converts SystemCallError to Error with pathname-first message.
  def self.stat_file(pathname)
    File.stat(pathname)
  rescue SystemCallError => e
    raise Error.new("#{pathname}: #{SystemCallError.new(e.errno)}")
  end

  def self.first_row_is_header?(csv)
    !csv.empty? && header?(csv[0])
  end

  def self.header?(csv_row)
    csv_row.size > 1 && csv_row[1] == CSV_PIN_COLUMN_NAME
  end

  def self.read_container_name_order(csv_pathname_or_io)
    result = {}
    io =
      if csv_pathname_or_io.is_a?(String)
        open_file(csv_pathname_or_io, 'r')
      else
        csv_pathname_or_io
      end
    begin
      csv = CSV.new(io)
      row = csv.shift
      return result if row.nil? || row.empty?
      index = 0
      if !header?(row)
        result[row[0]] = index
        index += 1
      end
      loop do
        row = csv.shift
        break unless row
        result[row[0]] = index
        index += 1
      end
      result
    ensure
      io.close if !io.closed? && csv_pathname_or_io.is_a?(String)
    end
  end

  def initialize(**args)
    super
    @options = nil
    @exit_status = 0
  end

  def usage_message
    "usage: #{@prog_name} [-h] sort [-hiv] [-r <file>] [<file> ...]"
  end

  def update_exit_status(status)
    @exit_status = status if @exit_status == 0
  end

  def parse_sort_options
    @options = SortOptions.new
    @options.help = false
    @options.in_place = false
    @options.verbose = false
    @options.ref_file = nil
    begin
      OptionParser.new do |opts|
        opts.on('-h', '--help') { @options.help = true }
        opts.on('-i') { @options.in_place = true }
        opts.on('-v') { @options.verbose = true }
        opts.on('-r FILE') { |arg| @options.ref_file = arg }
      end.parse!(@argv)
    rescue OptionParser::ParseError => e
      warnx(e)
      usage
    end
  end

  def create_in_place_file_io_pair(pathname)
    f_in = self.class.open_file(pathname, 'r')
    begin
      f_out = nil
      f_out = self.class.open_file(
        "#{pathname}.tmp",
        File::WRONLY | File::CREAT | File::EXCL
      )
      IOPair.new(f_in, f_out)
    rescue StandardError => e
      f_in.close
      f_out.close if f_out
      raise
    end
  end

  def sort(io_pair, container_name_order)
    csv = CSV.new(io_pair.in).read
    return if csv.empty?
    if csv[0].size < 2
      pathname = io_pair.in.respond_to?(:path) ? io_pair.in.path : '-'
      raise Error.new("#{pathname}: Too few columns in shipment CSV")
    end
    csv_header = self.class.first_row_is_header?(csv) ? csv.shift : nil
    # Ruby's sort is not stable; use sort_by with n to make it stable.
    if container_name_order
      # Sort by container name order, n, and then pin position.
      per_container_name_n = 0
      container_name_n = {}
      csv.sort_by! do |x|
        n =
          if container_name_n.include?(x[0])
            container_name_n[x[0]]
          else
            per_container_name_n += 1
            container_name_n[x[0]] = per_container_name_n
            per_container_name_n
          end
        [container_name_order[x[0]], n, x[1].to_i]
      end
    else
      # Sort by container name, pin position, and then n.
      n = 0
      csv.sort_by! do |x|
        n += 1
        [x[0], x[1].to_i, n]
      end
    end
    csv_out = CSV.new(io_pair.out)
    csv_out << csv_header if csv_header
    csv.each { |row| csv_out << row }
  end

  # Sets the user ID, group ID, and file mode from the source file stat to the
  # destination file at the specified pathname, as allowed by permissions.
  #
  # See also:
  # * <https://man.netbsd.org/cp.1>
  # * <https://man.netbsd.org/chmod.2>
  # * <https://nxr.netbsd.org/xref/src/bin/cp/utils.c#setfile>
  def set_file_attributes(fs_src, fp_dst)
    # Keep only S_ISUID | S_ISGID | S_IRWXU | S_IRWXG | S_IRWXO.
    mode = fs_src.mode & 06777
    begin
      File.chown(fs_src.uid, fs_src.gid, fp_dst)
    rescue SystemCallError => e
      if !defined?(Errno::EPERM) || !e.is_a?(Errno::EPERM)
        warnx("#{fp_dst}: chown: #{SystemCallError.new(e.errno)}")
        update_exit_status(1)
      end
      # Drop S_ISUID | S_ISGID.
      mode &= ~06000
    end
    begin
      File.chmod(mode, fp_dst)
    rescue SystemCallError => e
      warnx("#{fp_dst}: chmod: #{SystemCallError.new(e.errno)}")
      update_exit_status(1)
    end
  end

  def run
    usage if !@argv.empty? && ['-h', '--help'].include?(@argv[0])
    command = @argv.shift
    if !command
      warnx("Command required")
      usage
    end
    case command
    when 'sort'
      run_sort
    else
      warnx("#{command}: Invalid command")
      usage
    end
  end

  def run_sort
    parse_sort_options
    usage if @options.help
    container_name_order = nil
    if @options.ref_file
      begin
        container_name_order = self.class.read_container_name_order(
          @options.ref_file
        )
      rescue Error => e
        errx(1, e.to_s)
      rescue SystemCallError => e
        errx(1, "#{@options.ref_file}: #{SystemCallError.new(e.errno)}")
      end
    end
    @argv << '-' if @argv.empty?
    @argv.each do |pathname|
      begin
        if pathname == '-'
          sort(IOPair.new(@stdin, @stdout), container_name_order)
        elsif @options.in_place
          iop = create_in_place_file_io_pair(pathname)
          begin
            @stdout.puts(pathname) if @options.verbose
            sort(iop, container_name_order)
            iop.in.close
            iop.out.close
            set_file_attributes(
              self.class.stat_file(iop.in.path),
              iop.out.path
            )
            FileUtils.mv(iop.out.path, iop.in.path, force: true)
          ensure
            iop.in.close unless iop.in.closed?
            iop.out.close unless iop.out.closed?
            FileUtils.rm_f(iop.out.path)
          end
        else
          iop = IOPair.new(self.class.open_file(pathname, 'r'), @stdout)
          begin
            @stdout.puts(pathname) if @options.verbose
            sort(iop, container_name_order)
          ensure
            iop.in.close unless iop.in.closed?
          end
        end
      rescue Error, SystemCallError => e
        warnx(e.to_s)
        update_exit_status(1)
      end
    end
    @exit_status
  end
end

exit(IMCAShipmentToolApp.new.run) if __FILE__ == $0
